import torch
import torch.nn as nn
import torch.nn.functional as F
from nltk.translate.bleu_score import sentence_bleu, SmoothingFunction
from typing import Optional, Literal, TypeAlias, Union
from ..training.utils import TokenizedDict
from ..metrics.tensorbleu import tensor_sentence_bleu


SimpleMrqBenchScoreDict: TypeAlias = dict[
    Literal['bleu', 'cosine', 'length', 'raw', 'score'], dict[
        Literal['all', 'mean', 'max', 'min'], torch.Tensor
    ]
]

class SimpleMrqBenchScore:
    def __init__(
            self,
            shared_embedding: nn.Embedding,
            bleu_mode: Literal['separate', 'combined'] = 'separate',
            bleu_factor: float = 0.5,
            bleu_ref_factor: float = 0.5,
            bleu_saved_factor: float = 0.5,
            bleu_first_ref_factor: Optional[float] = None,
            bleu_first_saved_factor: Optional[float] = None,
            cos_factor: float = 0.5,
            cos_ref_factor: float = 0.5,
            cos_saved_factor: float = 0.5,
            multi_cos_ref_factor: float = 0.3,
            multi_cos_saved_factor: float = 0.5,
            multi_cos_running_mean_factor: float = 0.2,
            allow_not_summing_factors: bool = False,
            neg_reward_len: bool = False,
            max_rewarded_len: int = None,
            target_len_as_ref: bool = False,
            len_factor: int = None,
            running_mean_decay: float = 0.2,
            bleu_saved_weights: tuple = (0.5, 0.5),
            bleu_ref_weights: tuple = (0.5, 0.5),
            rewards_scale: float = 1.0,
            debug_mode: int = 0,
            use_tensor_bleu: bool = False,
            tensor_bleu_smoothing: Literal['none', 'floor', 'add-k', 'exp'] = 'exp',
    ):
        self.shared_embedding = shared_embedding
        self.bleu_mode = bleu_mode

        self.bleu_factor = bleu_factor
        self.bleu_ref_factor = bleu_ref_factor
        self.bleu_saved_factor = bleu_saved_factor
        self.bleu_first_ref_factor = bleu_first_ref_factor if bleu_first_ref_factor is not None else bleu_ref_factor
        self.bleu_first_saved_factor = bleu_first_saved_factor if bleu_first_saved_factor is not None else bleu_saved_factor
        self.cos_factor = cos_factor
        self.cos_ref_factor = cos_ref_factor
        self.cos_saved_factor = cos_saved_factor
        self.multi_cos_ref_factor = multi_cos_ref_factor
        self.multi_cos_saved_factor = multi_cos_saved_factor
        self.multi_cos_running_mean_factor = multi_cos_running_mean_factor
        self.neg_reward_len = neg_reward_len
        self.max_rewarded_len = max_rewarded_len
        self.target_len_as_ref = target_len_as_ref
        self.len_factor = len_factor
        self.running_mean_decay = running_mean_decay
        self.bleu_ref_weights = bleu_ref_weights
        self.bleu_saved_weights = bleu_saved_weights
        self.rewards_scale = rewards_scale
        self.bleu_smoothing = SmoothingFunction().method4
        self.debug_mode = debug_mode
        self.use_tensor_bleu = use_tensor_bleu
        self.tensor_bleu_smoothing = tensor_bleu_smoothing

        self.prev_data_running_mean = None

        if not allow_not_summing_factors:
            assert self.bleu_factor + self.cos_factor + self.len_factor == 1.0
            assert self.multi_cos_ref_factor + self.multi_cos_saved_factor + self.multi_cos_running_mean_factor == 1.0
            assert self.bleu_ref_factor + self.bleu_saved_factor == 1.0
            assert self.cos_ref_factor + self.cos_saved_factor == 1.0


    def _sentence_bleu(self, input_ids: tuple[torch.Tensor, torch.Tensor, torch.Tensor],
                       masks: tuple[torch.Tensor, torch.Tensor, torch.Tensor], is_first_step: bool = False) -> float:
        generated, reference, saved_data = input_ids
        generated_mask, reference_mask, saved_data_mask = masks

        generated = generated.tolist()[:generated_mask.sum().item()]
        reference = reference.tolist()[:reference_mask.sum().item()]
        saved_data = saved_data.tolist()[:saved_data_mask.sum().item()]

        if self.bleu_mode == 'separate':
            ref_bleu = sentence_bleu([reference], generated, weights=self.bleu_ref_weights,
                                     smoothing_function=self.bleu_smoothing)
            saved_bleu = sentence_bleu([saved_data], generated, weights=self.bleu_saved_weights,
                                       smoothing_function=self.bleu_smoothing)

            if is_first_step:
                return self.bleu_first_ref_factor * ref_bleu + self.bleu_first_saved_factor * saved_bleu
            else:
                return self.bleu_ref_factor * ref_bleu + self.bleu_saved_factor * saved_bleu
        else:
            return sentence_bleu(
                [reference, saved_data], generated,
                weights=self.bleu_ref_weights, smoothing_function=self.bleu_smoothing
            )

    def batch_bleu(self, generated: TokenizedDict, reference: TokenizedDict, saved_data: TokenizedDict, is_first_step: bool = False) -> Union[list[float], torch.Tensor]:
        if self.use_tensor_bleu:
          return self.batch_tensor_bleu(generated, reference, saved_data, is_first_step)

        batch_size = generated['input_ids'].size(0)

        return [
            self._sentence_bleu(
                input_ids=(generated['input_ids'][i], reference['input_ids'][i], saved_data['input_ids'][i]),
                masks=(generated['attention_mask'][i], reference['attention_mask'][i], saved_data['attention_mask'][i]),
                is_first_step=is_first_step,
            ) for i in range(batch_size)
        ]

    def batch_tensor_bleu(self, generated: TokenizedDict, reference: TokenizedDict, saved_data: TokenizedDict, is_first_step: bool = False) -> torch.Tensor:
        if self.bleu_mode == 'separate':
            ref_bleu = tensor_sentence_bleu(
                generated['input_ids'], reference['input_ids'].unsqueeze(1),
                weights=torch.tensor(self.bleu_ref_weights), smoothing_method=self.tensor_bleu_smoothing
            )
            saved_bleu = tensor_sentence_bleu(
                generated['input_ids'], saved_data['input_ids'].unsqueeze(1),
                weights=torch.tensor(self.bleu_saved_weights), smoothing_method=self.tensor_bleu_smoothing
            )

            if is_first_step:
                return self.bleu_first_ref_factor * ref_bleu + self.bleu_first_saved_factor * saved_bleu
            else:
                return self.bleu_ref_factor * ref_bleu + self.bleu_saved_factor * saved_bleu
        else:
            device = generated['input_ids'].device
            return tensor_sentence_bleu(
                generated['input_ids'], torch.tensor([reference['input_ids'], saved_data['input_ids']]).to(device),
                weights=torch.tensor(self.bleu_ref_weights), smoothing_method=self.tensor_bleu_smoothing
            )

    def _sequence_embedding(self, sequence: TokenizedDict) -> torch.Tensor:
        input_ids = sequence['input_ids']
        attention_mask = sequence['attention_mask']

        # Get embeddings
        embeddings = self.shared_embedding(input_ids)

        # Apply attention mask
        mask_expanded = attention_mask.unsqueeze(-1)
        masked_embeddings = embeddings * mask_expanded

        # Compute mean with masking
        sum_embeddings = torch.sum(masked_embeddings, dim=1)
        token_counts = torch.sum(mask_expanded, dim=1)
        token_counts = torch.clamp(token_counts, min=1e-8)  # Avoid division by zero

        return sum_embeddings / token_counts

    def _cosine_sim(self, generated: TokenizedDict, reference: TokenizedDict, saved_data: TokenizedDict):
        generated_emb = F.normalize(self._sequence_embedding(generated), dim=-1)
        saved_data_emb = F.normalize(self._sequence_embedding(saved_data), dim=-1)
        reference_emb = F.normalize(self._sequence_embedding(reference), dim=-1)

        gen_and_saved = F.cosine_similarity(generated_emb, saved_data_emb, dim=1)
        gen_and_ref = F.cosine_similarity(generated_emb, reference_emb, dim=1)

        return torch.clamp(gen_and_saved, min=0), torch.clamp(gen_and_ref, min=0)

    def _cosine_sim_running_mean(self, generated: TokenizedDict, reference: TokenizedDict, saved_data: TokenizedDict):
        generated_emb = F.normalize(self._sequence_embedding(generated), dim=-1)
        saved_data_emb = F.normalize(self._sequence_embedding(saved_data), dim=-1)
        reference_emb = F.normalize(self._sequence_embedding(reference), dim=-1)
        running_emb = F.normalize(self.prev_data_running_mean, dim=-1)

        gen_and_saved = F.cosine_similarity(generated_emb, saved_data_emb, dim=1)
        gen_and_ref = F.cosine_similarity(generated_emb, reference_emb, dim=1)
        gen_and_mean = F.cosine_similarity(generated_emb, running_emb, dim=1)

        return torch.clamp(gen_and_saved, min=0), torch.clamp(gen_and_ref, min=0), torch.clamp(gen_and_mean, min=0)

    def batch_cosine(self, generated: TokenizedDict, reference: TokenizedDict, saved_data: TokenizedDict,
                     include_running_mean: bool = False) -> torch.Tensor:
        if include_running_mean:
            gen_and_saved, gen_and_ref, gen_and_mean = self._cosine_sim_running_mean(generated, reference, saved_data)
            return self.multi_cos_saved_factor * gen_and_saved + self.multi_cos_ref_factor * gen_and_ref + self.multi_cos_running_mean_factor * gen_and_mean
        else:
            gen_and_saved, gen_and_ref = self._cosine_sim(generated, reference, saved_data)
            return self.cos_saved_factor * gen_and_saved + self.cos_ref_factor * gen_and_ref

    def len_reward(self, generated: TokenizedDict, reference: TokenizedDict) -> torch.Tensor:
        target_lens = reference['attention_mask'].sum(dim=-1) if self.target_len_as_ref else self.max_rewarded_len
        lens = generated['attention_mask'].sum(dim=-1)
        neg_lens = target_lens / lens if self.neg_reward_len else 1.0
        len_reward = torch.where(lens >= target_lens, neg_lens, lens / target_lens)
        return len_reward

    def reset_running_mean(self):
        self.prev_data_running_mean = None

    def init_running_mean(self, prev_data: TokenizedDict):
        self.prev_data_running_mean = self._sequence_embedding(prev_data)

    def update_running_mean(self, prev_data: TokenizedDict):
        self.prev_data_running_mean = (1 - self.running_mean_decay) * self._sequence_embedding(
            prev_data) + self.running_mean_decay * self.prev_data_running_mean

    def __call__(
            self,
            generated: TokenizedDict,
            reference: TokenizedDict,
            saved_data: TokenizedDict,
            prev_data: TokenizedDict = None,
    ) -> SimpleMrqBenchScoreDict:
        if prev_data is not None:
            if self.prev_data_running_mean is None:
                self.init_running_mean(prev_data)
            else:
                self.update_running_mean(prev_data)

        device = generated['input_ids'].device

        bleu = self.batch_bleu(generated, reference, saved_data, is_first_step=prev_data is None)
        cosine = self.batch_cosine(generated, reference, saved_data, include_running_mean=prev_data is not None)
        len_reward = self.len_reward(generated, reference)

        tensor_bleu = torch.tensor(bleu, device=device) if not self.use_tensor_bleu else bleu

        raw_score = self.bleu_factor * tensor_bleu + self.cos_factor * cosine + self.len_factor * len_reward
        scaled_score = raw_score * self.rewards_scale

        return {
            'bleu': {
                'all': tensor_bleu,
                'mean': tensor_bleu.mean(),
                'max': tensor_bleu.max(),
                'min': tensor_bleu.min(),
            },
            'cosine': {
                'all': cosine,
                'mean': cosine.mean(),
                'max': cosine.max(),
                'min': cosine.min(),
            },
            'length': {
                'all': len_reward,
                'mean': len_reward.mean(),
                'max': len_reward.max(),
                'min': len_reward.min(),
            },
            'raw': {
                'all': raw_score,
                'mean': raw_score.mean(),
                'max': raw_score.max().item(),
                'min': raw_score.min().item(),
            },
            'score': {
                'all': scaled_score,
                'mean': scaled_score.mean(),
                'max': scaled_score.max(),
                'min': scaled_score.min(),
            }
        }



import torch
import torch.nn as nn
from typing import TypedDict

class AdaptivePositionalMemoryNorm(nn.Module):
    def __init__(
        self,
        num_slots: int,
        dim: int,
        decay: float = 0.9,
        use_scale: bool = True,
        use_gate: bool = True,
        init_gate: float = -2.0,
        per_dim_scale: bool = False,
    ):
        super(AdaptivePositionalMemoryNorm, self).__init__()
        self.use_gate = use_gate
        self.num_slots = num_slots
        self.dim = dim
        self.decay = decay
        self.eps = 1e-6

        # Learnable parameters
        scale_shape = (num_slots, 1) if not per_dim_scale else (dim,)
        self.scale = nn.Parameter(torch.ones(*scale_shape)) if use_scale else None
        self.gate = nn.Parameter(torch.full((num_slots, 1), init_gate)) if use_gate else None

        # EMA buffers
        self.register_buffer("ema_rms", torch.ones(num_slots, 1))

        # Initialize parameters
        if self.scale is not None:
            nn.init.normal_(self.scale, mean=1.0, std=0.1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Calculate current RMS per slot
        # x: [batch_size, num_slots, dim]
        current_rms = x.pow(2).mean(dim=-1, keepdim=True).sqrt()  # [batch, num_slots, 1]
        slot_rms = current_rms.mean(dim=0)  # [num_slots, 1] (average over batch)

        # Update EMA during training
        if self.training:
            self.ema_rms = self.decay * self.ema_rms + (1 - self.decay) * slot_rms.detach() # [num_slots, 1]

        # Normalize using EMA statistics
        x_norm = x * torch.rsqrt(self.ema_rms + self.eps) # [batch_size, num_slots, dim] * [num_slots, 1]

        # Apply learned scale per slot
        if self.scale is not None:
            x_norm = x_norm * self.scale # [batch_size, num_slots, dim] * [num_slots, 1] or [dim]

        # Apply gating mechanism
        if self.use_gate:
            gate = torch.sigmoid(self.gate)  # [num_slots, 1]
            return gate * x_norm + (1 - gate) * x # [batch_size, num_slots, dim] * [num_slots, 1]

        return x_norm

class AdaptiveRMSMemoryNorm(nn.Module):
    def __init__(
        self,
        dim: int,
        use_gate: bool = True,
        decay: float = 0.99,
        init_scale: float = 1.0,
        init_gate: float = -4.0  # Start with gate closed (no normalization)
    ):
        super().__init__()
        self.use_gate = use_gate
        self.scale = nn.Parameter(torch.ones(dim) * init_scale)
        self.gate = nn.Parameter(torch.tensor([init_gate]))  # Scalar gate for this layer
        self.eps = 1e-6
        self.decay = decay
        self.register_buffer("ema_rms", torch.ones(1))  # Scalar EMA RMS for the entire layer's STM

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x shape: [batch_size, num_slots, dim]
        if self.training and hasattr(self, 'ema_rms'):
            # Compute current RMS across all slots and batch (scalar)
            current_rms = x.pow(2).mean(dim=-1).mean().sqrt()
            self.ema_rms = self.ema_rms * self.decay + current_rms * (1 - self.decay)
            rms = self.ema_rms
        else:
            # Compute RMS per slot (mean over dim)
            rms = x.pow(2).mean(-1, keepdim=True).sqrt()  # [batch_size, num_slots, 1]

        # Normalize each slot's embedding vector
        normalized = x * torch.rsqrt(rms + self.eps)
        normalized = normalized * self.scale  # Apply per-dimension scaling

        if self.use_gate:
            gate_factor = torch.sigmoid(self.gate)  # Scalar gate (0-1)
            return normalized * gate_factor + x * (1 - gate_factor)
        else:
            return normalized

class SimpleRMSMemoryNorm(nn.Module):
    def __init__(
        self,
        dim: int,
        use_gate: bool = True,
        init_scale: float = 1.0,
        init_gate: float = -4.0
    ):
        super().__init__()
        self.use_gate = use_gate
        self.scale = nn.Parameter(torch.ones(dim) * init_scale)
        self.gate = nn.Parameter(torch.tensor([init_gate]))  # Scalar gate
        self.eps = 1e-6

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        rms = x.pow(2).mean(-1, keepdim=True).sqrt()  # [batch_size, num_slots, 1]
        normalized = x * torch.rsqrt(rms + self.eps)
        normalized = normalized * self.scale  # Apply per-dimension scaling

        if self.use_gate:
            gate_factor = torch.sigmoid(self.gate)  # Scalar gate (0-1)
            return normalized * gate_factor + x * (1 - gate_factor)
        else:
            return normalized

class MemoryLayerNorm(nn.Module):
    def __init__(
        self,
        dim: int,
        use_gate: bool = True,
        init_scale: float = 1.0,
        init_gate: float = -4.0  # Start with gate closed (no normalization)
    ):
        super().__init__()
        self.use_gate = use_gate
        self.norm = nn.LayerNorm(dim)  # Normalizes across embedding dimensions per slot
        self.gate = nn.Parameter(torch.tensor([init_gate]))  # Scalar gate for this layer
        self.scale = nn.Parameter(torch.ones(dim) * init_scale)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        normalized = self.norm(x)  # Apply LayerNorm across embedding dimensions (per slot)
        normalized = normalized * self.scale  # Per-dimension scaling

        if self.use_gate:
            gate_factor = torch.sigmoid(self.gate)  # Scalar gate (0-1)
            return normalized * gate_factor + x * (1 - gate_factor)
        else:
            return normalized

class MemoryNormConfig(TypedDict):
    num_slots: int
    decay: float
    use_scale: bool
    use_gate: bool
    init_gate: float
    init_scale: float
    per_dim_scale: bool

def init_memory_norm(
    norm_type: str,
    dim: int,
    num_slots: int = None,
    decay: float = 0.9,
    use_scale: bool = True,
    use_gate: bool = True,
    init_gate: float = -2.0,
    init_scale: float = 1.0,
    per_dim_scale: bool = False,
) -> nn.Module:
    assert norm_type in ['layer', 'rms', 'adaptive', 'positional', 'classic-rms']
    if norm_type == 'layer':
        return MemoryLayerNorm(dim, use_gate, init_scale, init_gate)
    elif norm_type == 'rms':
        return SimpleRMSMemoryNorm(dim, use_gate, init_scale, init_gate)
    elif norm_type == 'adaptive':
        return AdaptiveRMSMemoryNorm(dim, use_gate, decay, init_scale, init_gate)
    elif norm_type == 'positional':
        return AdaptivePositionalMemoryNorm(num_slots, dim, decay, use_scale, use_gate, init_gate, per_dim_scale)
    elif norm_type == 'classic-rms':
        return nn.RMSNorm(dim)
    return MemoryLayerNorm(dim, use_gate, init_scale, init_gate)



import torch
import torch.nn as nn
import torch.nn.functional as F
from ..transformers.moe import MoeRouter

class MoeFeedForwardVectorized(nn.Module):
    """
    Vectorized MoE - current implementation is incorrect - it calculates all the experts, then selects the correct ones.

    Commented out implementation is fixing this problem, but is causing memory overflows, because of experts weights
    indexing - it's using ~15x more memory, than dense model of similar size, so it's currently not viable.

    It's recommended to use standard MoE from rxlm.transformers.moe instead.
    """

    def __init__(
            self,
            embed_dim: int,
            hidden_dim: int,
            num_experts: int,
            activation: nn.Module,
            top_k: int = 1,
            dropout: float = 0.0,
            *args,
            **kwargs
    ):
        super(MoeFeedForwardVectorized, self).__init__(*args, **kwargs)
        self.embed_dim = embed_dim
        self.num_experts = num_experts
        self.top_k = top_k

        self.router = MoeRouter(embed_dim, num_experts, top_k)

        # Batch all expert parameters together
        self.w1 = nn.Parameter(torch.empty(num_experts, embed_dim, self._w1_dim_factor(hidden_dim)))
        self.b1 = nn.Parameter(torch.zeros(num_experts, self._w1_dim_factor(hidden_dim)))
        self.w2 = nn.Parameter(torch.empty(num_experts, hidden_dim, embed_dim))
        self.b2 = nn.Parameter(torch.zeros(num_experts, embed_dim))
        self.activation = activation
        self.dropout = nn.Dropout(dropout)

        # Initialize parameters
        self._init_linear_parameters()
        nn.init.zeros_(self.b1)
        nn.init.zeros_(self.b2)

    def _init_linear_parameters(self):
        nn.init.kaiming_normal_(self.w1, nonlinearity='relu')
        nn.init.kaiming_normal_(self.w2, nonlinearity='relu')

    def _w1_dim_factor(self, hidden_dim: int) -> int:
        return hidden_dim

    def _activate(self, h: torch.Tensor):
        return self.activation(h)

    def router_loss(self):
        return self.router.aux_loss

    def forward(self, x: torch.Tensor):
        orig_shape = x.shape
        x = x.view(-1, self.embed_dim)  # [batch*seq_len, embed_dim]

        # Get routing weights and indices
        weights, indices = self.router(x)  # [batch*seq_len, top_k]

        # Create expert masks and combine it with masks
        mask = F.one_hot(indices, self.num_experts).float()  # [batch*seq_len, top_k, num_experts]
        weights = (weights.unsqueeze(-1) * mask).sum(dim=1)  # [batch*seq_len, num_experts]

        # Expert computation
        x = x.unsqueeze(1).expand(-1, self.num_experts, -1)  # [batch*seq_len, num_experts, embed_dim]

        # First linear layer
        h = torch.einsum('bie,ieh->bih', x, self.w1) + self.b1  # [batch*seq_len, num_experts, hidden_dim]
        h = self._activate(h)
        h = self.dropout(h)

        # Second linear layer (projection back to embed_dim)
        out = torch.einsum('bih,ihe->bie', h, self.w2) + self.b2  # [batch*seq_len, num_experts, embed_dim]

        # Weighted sum of expert outputs
        out = (out * weights.unsqueeze(-1)).sum(dim=1)  # [batch*seq_len, embed_dim]

        return out.view(*orig_shape)
        # orig_shape = x.shape
        # x = x.view(-1, self.embed_dim)  # [batch*seq_len, embed_dim]
        #
        # # Get routing weights and indices
        # weights, indices = self.router(x)  # [B*T, top_k], [B*T, top_k]
        #
        # # Flatten indices and weights
        # batch_size = x.shape[0]
        # top_k = indices.shape[1]
        # indices_flat = indices.view(-1)  # [B*T * top_k]
        #
        # # Compute contributions for selected experts without materializing large tensors
        # # First Layer:
        # # Compute all expert contributions first (but this may still be memory-heavy)
        # # Alternative: Compute contributions for selected experts directly
        # # ... (see detailed steps below)
        #
        # # Alternative approach using gather and batched operations
        # x_expanded = x.unsqueeze(1).repeat(1, top_k, 1).view(-1, self.embed_dim)  # [B*T*top_k, D]
        #
        # # Compute first layer contributions using gather
        # # indices_flat has shape [B*T*top_k]
        # # selected_w1 is self.w1[indices_flat], but we compute the product inline
        # h = torch.einsum(
        #     'be, eih -> bh',
        #     x_expanded,
        #     self.w1[indices_flat]
        # ) + self.b1[indices_flat]
        # h = self._activate(h)
        # h = self.dropout(h)
        #
        # # Second layer:
        # out = torch.einsum(
        #     'bh, eho -> beo',
        #     h,
        #     self.w2[indices_flat]
        # ).squeeze(-1) + self.b2[indices_flat]
        #
        # # Reshape and apply weights
        # out = out.view(batch_size, top_k, -1)
        # weights = weights.view(batch_size, top_k, 1)
        # out = (out * weights).sum(dim=1)
        #
        # return out.view(*orig_shape)


class GatedMoeFeedForwardVectorized(MoeFeedForwardVectorized):
    """Gated Mixture-of-Experts Feed-Forward layer - enable GLU-based activations for MoE"""

    def __init__(
            self,
            embed_dim: int,
            hidden_dim: int,
            num_experts: int,
            activation: nn.Module = nn.SiLU(),
            top_k: int = 1,
            dropout: float = 0.1,
            *args,
            **kwargs
    ):
        super(GatedMoeFeedForwardVectorized, self).__init__(
            embed_dim=embed_dim,
            hidden_dim=hidden_dim,
            num_experts=num_experts,
            activation=activation,
            top_k=top_k,
            dropout=dropout,
            *args,
            **kwargs
        )

    def _init_linear_parameters(self):
        nn.init.kaiming_normal_(self.w1, nonlinearity='relu')
        nn.init.kaiming_normal_(self.w2, nonlinearity='linear')

    def _w1_dim_factor(self, hidden_dim: int) -> int:
        return 2 * hidden_dim

    def _activate(self, h: torch.Tensor):
        a, b = h.chunk(2, dim=-1)
        return a * self.activation(b)



import torch
import torch.nn.functional as F
from torch.utils.data import DataLoader
from typing import Optional, TypedDict, Union, Literal
from ..llm.models import DecoderOnlyTransformer
from ..training.utils import smart_concat, TokenizedDict
from .score import SimpleMrqBenchScore, SimpleMrqBenchScoreDict
from .models import RxTMrqBenchModel, RxTMrqBenchAction
from .dataset import RxlmMrqBenchDataset, LlmMrqBenchDataset
from .sampler import MrqBenchBatchSampler

class MrqBenchTokenizerConfig(TypedDict):
    pad_token_id: int
    end_token_id: int
    answer_token_id: int


class SamplerConfig(TypedDict):
    temperature: float
    top_k: Optional[int]
    top_p: Optional[float]


class MrqBenchRunner:
    def __init__(
            self,
            device: torch.device,
            score: SimpleMrqBenchScore,
            max_seq_len: int,
            interaction_len: int,
            tokenizer_config: MrqBenchTokenizerConfig,
            sampler_config: Optional[SamplerConfig] = None,
            use_amp: bool = False,
            dtype: torch.dtype = torch.float32,
            teacher_forcing: bool = True,
    ):
        self.score = score
        self.device = device

        # Batch Sampler for answer generation
        self.generator = None
        self.sampler_config = SamplerConfig(
            temperature=1.0,
            top_k=None,
            top_p=None,
        ) if sampler_config is None else sampler_config

        self.pad_token_id = tokenizer_config.get('pad_token_id', 0)
        self.end_token_id = tokenizer_config.get('end_token_id', 3)
        self.answer_token_id = tokenizer_config.get('answer_token_id', 6)

        self.use_amp = use_amp
        self.dtype = dtype

        self.max_seq_len = max_seq_len
        self.interaction_len = interaction_len

        # Dynamic fields, updated for each curriculum step
        self.dataset = None
        self.teacher_forcing = teacher_forcing
        self.model = None


    def encode_and_update_stm(self, query: TokenizedDict, answer: TokenizedDict):
        """Encode interaction and update STM."""
        # 1. Encode data and update memory - with autocast on/off
        if self.use_amp:
            with torch.amp.autocast(device_type=self.device.type, dtype=self.dtype):
                # 2. Concatenate batch of queries and answers (they are already on training device)
                inputs = smart_concat(query, answer, self.interaction_len, self.pad_token_id)
                # 3. Encode data and update STM
                self.model(inputs['input_ids'], attention_mask=inputs['attention_mask'], action=RxTMrqBenchAction.UPDATE)
        else:
            # 2. Concatenate batch of queries and answers (they are already on training device)
            inputs = smart_concat(query, answer, self.interaction_len, self.pad_token_id)
            # 3. Encode data and update STM
            self.model(inputs['input_ids'], attention_mask=inputs['attention_mask'], action=RxTMrqBenchAction.UPDATE)

    def generate_answer(self, query: TokenizedDict, timing_log: list[dict[str, float]]) -> TokenizedDict:
        """Generate response using batch sampler with decoder."""
        # 1. Generate answer with BatchSampler - with autocast on/off
        if self.use_amp:
            with torch.amp.autocast(device_type=self.device.type, dtype=self.dtype):
                input_ids, attention_mask, _ = self.generator(
                    query['input_ids'],
                    query['attention_mask'],
                    max_gen_len=self.interaction_len,
                    dtype=self.dtype,
                    timing_log=timing_log,
                    **self.sampler_config,
                )
        else:
            input_ids, attention_mask, _ = self.generator(
                query['input_ids'],
                query['attention_mask'],
                max_gen_len=self.interaction_len,
                dtype=self.dtype,
                timing_log=timing_log,
                **self.sampler_config,
            )
        # 2. Convert generated answer to TokenizedDict
        generated_answer: TokenizedDict = {
            'input_ids': input_ids,
            'attention_mask': attention_mask,
        }

        return generated_answer

    def compute_score(
            self, generated: TokenizedDict, reference: TokenizedDict, saved_data: tuple[TokenizedDict, TokenizedDict],
            prev_data: tuple[TokenizedDict, TokenizedDict] = None
    ) -> SimpleMrqBenchScoreDict:
        """Compute reward based on memory retention (e.g., BLEU-4)."""
        # 1. Move sequences to GPU for reward calculation
        saved_query, saved_answer = self._move_multiple_batches(*saved_data)
        reference = self._move_batch(reference)
        prev_data = self._move_multiple_batches(*prev_data) if prev_data is not None else None

        saved_interaction = smart_concat(
            saved_query, saved_answer, max_length=self.interaction_len, pad_token_id=self.pad_token_id
        )
        prev_interaction = smart_concat(prev_data[0], prev_data[1], self.interaction_len,
                                 self.pad_token_id) if prev_data is not None else None

        # 2. Concat saved (previous) interaction and calculate reward using generated sequence, reference and saved data - with autocast on/off
        if self.use_amp:
            with torch.amp.autocast(device_type=self.device.type, dtype=self.dtype):
                scoring = self.score(generated, reference, saved_interaction, prev_data=prev_interaction)
        else:
            scoring = self.score(generated, reference, saved_interaction, prev_data=prev_interaction)

        # 4. Return scores for batch
        return scoring

    def _move_batch(self, batch: TokenizedDict) -> TokenizedDict:
        if self.use_amp:
            return {
                'input_ids': batch['input_ids'].to(self.device),
                'attention_mask': batch['attention_mask'].to(self.device),
            }
        else:
            return {
                'input_ids': batch['input_ids'].to(self.device, dtype=self.dtype),
                'attention_mask': batch['attention_mask'].to(self.device, dtype=self.dtype),
            }

    def _move_multiple_batches(self, *batches: TokenizedDict) -> list[TokenizedDict]:
        return [self._move_batch(batch) for batch in batches]

    def _cpu_detach(self, batch: TokenizedDict) -> TokenizedDict:
        return {
            'input_ids': batch['input_ids'].detach().cpu(),
            'attention_mask': batch['attention_mask'].detach().cpu(),
        }

    def _batch_detach(self, batch: TokenizedDict) -> TokenizedDict:
        return {
            'input_ids': batch['input_ids'].detach(),
            'attention_mask': batch['attention_mask'].detach(),
        }

    def _cpu_detach_multiple(self, *batches: TokenizedDict) -> list[TokenizedDict]:
        return [self._cpu_detach(batch) for batch in batches]

    def _rxlm_loader(self, batch_size: int):
        return DataLoader(
            self.dataset,
            batch_size=batch_size,
            shuffle=False,
            pin_memory=True,
            collate_fn=RxlmMrqBenchDataset.collate_mrl_batch,
        )

    def _llm_loader(self, batch_size: int):
        return DataLoader(
            self.dataset,
            batch_size=batch_size,
            shuffle=False,
            pin_memory=True,
            collate_fn=LlmMrqBenchDataset.collate_llm_batch,
        )

    def run_rxlm(self, batch_size: int, num_of_examples: int = None):
        # 1. Init evaluation DataLoader
        dataloader = self._rxlm_loader(batch_size)
        self.model.eval()

        results = []

        # 2. Run evaluation on all batch episodes
        for batch_idx, batch in enumerate(dataloader):
            with torch.no_grad():
                if num_of_examples is not None:
                    if batch_idx >= num_of_examples:
                        break
                if batch['query']['input_ids'].size(0) == batch_size:
                    # 3. Reset STM with random resets ratio and reward model running mean
                    self.model.reset_memory()
                    self.score.reset_running_mean()

                    # 4. Get batches for first queries, answers and all follow-up interactions
                    first_query, first_answer, interactions = batch['query'], batch['answer'], batch['interactions']
                    # 5. Encode and update STM with initial interactions (batch)
                    first_interaction = self._move_multiple_batches(first_query, first_answer)
                    self.encode_and_update_stm(*first_interaction)

                    # 6. Save follow-up interactions len and first query and answer as previous one for iteration
                    interactions_len = len(interactions)
                    query, answer = first_query, first_answer

                    prev_interaction = None

                    episode_results: list[SimpleMrqBenchScoreDict] = []
                    timing_log: list[dict[str, float]] = []

                    # 7. Run all follow-up interactions
                    for i, interaction in enumerate(interactions):
                        # 8. Generate batch of answers
                        next_query = self._move_batch(interaction['query'])
                        generated_answer = self.generate_answer(next_query, timing_log)

                        is_last_interaction = (i + 1) == interactions_len

                        detached_answer = self._batch_detach(generated_answer)

                        # 9. Depending on current strategy and step, compute reward
                        scoring = self.compute_score(
                            detached_answer, interaction['answer'], (query, answer), prev_data=prev_interaction
                        )

                        print(f'Step: {i + 1}. Mean reward: {scoring['score']['mean'].item()} / Min reward: {scoring['score']['min'].item()} / Max reward: {scoring['score']['max'].item()}')
                        print(f'Step: {i + 1}. Mean BLEU: {scoring['bleu']['mean'].item()} Mean Cosine: {scoring['cosine']['mean'].item()}')

                        episode_results.append(scoring)

                        # 10. Encode and update memory for the next interaction
                        if not is_last_interaction:
                            self.encode_and_update_stm(
                                next_query,
                                self._move_batch(interaction['answer']) if self.teacher_forcing else generated_answer
                            )

                        # 12. Save previous interaction
                        prev_interaction = (query, answer)
                        query, answer = interaction['query'], (interaction['answer'] if self.teacher_forcing else self._cpu_detach(generated_answer))

                    batch_mean_scores = torch.stack([item['score']['all'] for item in episode_results]).transpose(0, 1).mean(dim=-1)

                    episode = {
                        'scores': episode_results,
                        'mean_score': torch.stack([item['score']['mean'] for item in episode_results]).mean(),
                        'all': batch_mean_scores,
                        'mean': batch_mean_scores.mean(),
                        'max': batch_mean_scores.max(),
                        'min': batch_mean_scores.min(),
                        'timing': timing_log,
                    }
                    results.append(episode)

                    print(
                        f'Mean episode reward: {episode["mean_score"].item()} | {episode["mean"].item()} / Min episode reward: {episode["min"].item()} / Max episode reward: {episode["max"].item()}')

        # 15. Calculate average reward
        avg_score = torch.stack([result['mean_score'] for result in results]).mean() if len(results) > 0 else torch.tensor(0.0)

        return {
            'mean': avg_score.item(),
            'scores': results,
        }

    def run_llm(self, batch_size: int, num_of_examples: int = None):
        # 1. Init evaluation DataLoader
        dataloader = self._llm_loader(batch_size)
        self.model.eval()

        results = []

        # 2. Run evaluation on all batch episodes
        for batch_idx, batch in enumerate(dataloader):
            with torch.no_grad():
                if num_of_examples is not None:
                    if batch_idx >= num_of_examples:
                        break
                if batch['query']['input_ids'].size(0) == batch_size:
                    self.score.reset_running_mean()

                    # 4. Get batches for first queries, answers and all follow-up interactions
                    first_query, first_answer, interactions = batch['query'], batch['answer'], batch['interactions']

                    # 6. Save follow-up interactions len and first query and answer as previous one for iteration
                    query, answer = first_query, first_answer

                    prev_interaction = None

                    episode_results: list[SimpleMrqBenchScoreDict] = []
                    timing_log: list[dict[str, float]] = []

                    # 7. Run all follow-up interactions
                    for i, interaction in enumerate(interactions):
                        # 8. Generate batch of answers
                        next_context = self._move_batch(interaction['context'])
                        generated_answer = self.generate_answer(next_context, timing_log)

                        detached_answer = self._batch_detach(generated_answer)

                        # 9. Depending on current strategy and step, compute reward
                        scoring = self.compute_score(
                            detached_answer, interaction['answer'], (query, answer), prev_data=prev_interaction
                        )

                        print(f'Step: {i + 1}. Mean reward: {scoring['score']['mean'].item()} / Min reward: {scoring['score']['min'].item()} / Max reward: {scoring['score']['max'].item()}')
                        print(f'Step: {i + 1}. Mean BLEU: {scoring['bleu']['mean'].item()} Mean Cosine: {scoring['cosine']['mean'].item()}')

                        episode_results.append(scoring)
                        # 12. Save previous interaction
                        prev_interaction = (query, answer)
                        query, answer = interaction['query'], (
                            interaction['answer'] if self.teacher_forcing else self._cpu_detach(generated_answer))

                    batch_mean_scores = torch.stack([item['score']['all'] for item in episode_results]).transpose(0, 1).mean(dim=-1)

                    episode = {
                        'scores': episode_results,
                        'mean_score': torch.stack([item['score']['mean'] for item in episode_results]).mean(),
                        'all': batch_mean_scores,
                        'mean': batch_mean_scores.mean(),
                        'max': batch_mean_scores.max(),
                        'min': batch_mean_scores.min(),
                        'timing': timing_log,
                    }
                    results.append(episode)

                    print(f'Mean episode reward: {episode["mean_score"].item()} | {episode["mean"].item()} / Min episode reward: {episode["min"].item()} / Max episode reward: {episode["max"].item()}')


        # 15. Calculate average reward
        avg_score = torch.stack([result['mean_score'] for result in results]).mean() if len(results) > 0 else torch.tensor(0.0)

        return {
            'mean': avg_score.item(),
            'scores': results,
        }


    def __call__(self, mode: Literal['rxlm', 'llm'], model: Union[DecoderOnlyTransformer, RxTMrqBenchModel], dataset: Union[RxlmMrqBenchDataset, LlmMrqBenchDataset], batch_size: int, num_of_examples: int = None):
        self.model = model
        self.dataset = dataset

        self.generator = MrqBenchBatchSampler(
            self.model, self.device, end_token_id=self.end_token_id, answer_token_id=self.answer_token_id,
            pad_token_id=self.pad_token_id, use_self_attn_cache=True, use_stm_kv_cache=mode=='rxlm', use_first_empty_workaround=False
        )

        if mode == 'rxlm':
            scores = self.run_rxlm(batch_size, num_of_examples)
        else:
            scores = self.run_llm(batch_size, num_of_examples)

        return scores



import torch
from torch import nn
from typing import TypedDict, Optional, Iterator, Union, Literal
from huggingface_hub import PyTorchModelHubMixin
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast

from ..training.tokenizer import decode_post_process
from ..transformers.positional import RotaryPositionalEmbedding
from ..transformers.attention import init_attention
from ..transformers.llm_layers import ClassicTransformerLayer
from ..transformers.llm_models import ClassicTransformerDecoder
from ..transformers.ff import get_activation_layer
from ..transformers.sampler import sample
from ..utils import get_model_size
from ..experimental.attention import init_experimental_attention


class DecoderOnlyTransformerConfig(TypedDict):
    num_layers: int
    vocab_size: int
    embed_dim: int
    ff_dim: int
    att_heads: int
    seq_len: int
    use_flash_attention: bool
    use_gated: bool
    ff_activation: str
    ff_dropout: float
    att_dropout: float
    use_rms_norm: bool
    att_groups: int
    use_moe_ff: bool
    ff_num_experts: int
    ff_moe_top_k: int
    ff_num_shared_experts: int
    att_type: str
    att_num_experts: int
    att_num_query_experts: int
    att_num_query_groups: int
    att_num_global_tokens: int
    att_window_size: int
    use_head_norm: bool
    init_identity_norm: bool
    tie_embeddings: bool
    head_norm_type: str


class DecoderOnlyTransformer(nn.Module, PyTorchModelHubMixin, pipeline_tag="text-generation", license="apache-2.0"):
    """
    Research model for experiments with new attention layers.

    Currently, accepts SparseQueryAttention, GroupedMoeAttention, DeepMoeAttention and standard variants (MHA/GQA/MQA) for reference models
    """

    def __init__(
            self,
            num_layers: int = 6,
            vocab_size: int = 5000,
            embed_dim: int = 128,
            ff_dim: int = 384,
            att_heads: int = 16,
            seq_len: int = 256,
            use_flash_attention: bool = False,
            use_gated: bool = True,
            ff_activation: str = "swish",
            ff_dropout: float = 0.0,
            att_dropout: float = 0.0,
            use_rms_norm: bool = True,
            att_groups: int = 1,
            use_moe_ff: bool = False,
            ff_num_experts: int = 1,
            ff_moe_top_k: int = 1,
            ff_num_shared_experts: int = 0,
            num_initial_dense_layers: int = 0,
            dense_ff_dim: int = 384,
            att_type: str = 'sqa',
            att_num_experts: int = None,
            att_num_query_experts: int = None,
            att_num_query_groups: int = None,
            att_num_global_tokens: int = 16,
            att_window_size: int = 128,
            use_head_norm: bool = False,
            init_identity_norm: bool = False,
            tie_embeddings: bool = False,
            head_norm_type: str = 'layer_norm',
            router_amp: bool = False,
            router_dtype: torch.dtype = torch.float32,
            **kwargs
    ):
        super(DecoderOnlyTransformer, self).__init__(**kwargs)
        assert ff_activation in ['relu', 'gelu',
                                 'swish', 'silu', 'linear',
                                 'sigmoid'], 'Feed-forward activation could be "relu", "gelu", "swish", "silu", "linear", "sigmoid".'
        assert att_type in ['mha', 'gqa', 'mqa', 'gma', 'dma', 'sqa', 'flex', 'flex-sqa'], 'Self-attention type could be "mha", "gqa", "mqa", "gma", "dma", "sqa", "flex", "flex-sqa".'

        embedding = nn.Embedding(vocab_size, embed_dim)
        rope = RotaryPositionalEmbedding(embed_dim // att_heads, seq_len)

        ff_activation = get_activation_layer(ff_activation)

        if att_type in ['mha', 'gqa', 'mqa', 'sqa']:
            att_init = lambda: init_attention(embed_dim, att_heads, att_type, att_groups, rope=rope,
                                              use_flash_attention=use_flash_attention, dropout=att_dropout,
                                              max_seq_len=seq_len, is_causal=True, num_query_groups=att_num_query_groups)
        else:
            att_init = lambda: init_experimental_attention(embed_dim, att_heads, att_type, att_groups, rope=rope,
                                                           use_flash_attention=use_flash_attention, dropout=att_dropout,
                                                           max_seq_len=seq_len, is_causal=True, num_experts=att_num_experts,
                                                           num_query_experts=att_num_query_experts, num_query_groups=att_num_query_groups,
                                                           num_global_tokens=att_num_global_tokens, window_size=att_window_size)

        use_moe_att = att_type in ['gma', 'dma']

        def layer_init(i: int):
            if i < num_initial_dense_layers:
                return ClassicTransformerLayer(
                    embed_dim,
                    dense_ff_dim,
                    use_gated=use_gated,
                    use_moe=False,
                    ff_activation=ff_activation,
                    ff_dropout=ff_dropout,
                    use_rms_norm=use_rms_norm,
                    self_attention=att_init(),
                    use_moe_att=use_moe_att,
                )
            else:
                return ClassicTransformerLayer(
                    embed_dim,
                    ff_dim,
                    use_gated=use_gated,
                    use_moe=use_moe_ff,
                    num_experts=ff_num_experts,
                    num_shared_experts=ff_num_shared_experts,
                    moe_top_k=ff_moe_top_k,
                    ff_activation=ff_activation,
                    ff_dropout=ff_dropout,
                    use_rms_norm=use_rms_norm,
                    self_attention=att_init(),
                    use_moe_att=use_moe_att,
                    router_amp=router_amp,
                    router_dtype=router_dtype,
                )


        self.model = ClassicTransformerDecoder(
            embed_dim,
            vocab_size,
            embedding=embedding,
            layers=nn.ModuleList([layer_init(i) for i in range(num_layers)]),
            use_flash_attention=use_flash_attention,
            use_head_norm=use_head_norm,
            init_identity_norm=init_identity_norm,
            tie_embeddings=tie_embeddings,
            head_norm_type=head_norm_type,
        )

    def params_count(self):
        return get_model_size(self.model)

    def reset_self_attn_cache(self):
        return self.model.reset_self_attn_cache()

    def forward(self, x: torch.Tensor, attention_mask: torch.Tensor = None, use_self_attn_cache: bool = False, current_positions: torch.Tensor = None) -> torch.Tensor:
        return self.model(x, attention_mask=attention_mask, use_self_attn_cache=use_self_attn_cache, current_positions=current_positions)

    def _generate_single_token(
            self,
            input_ids: torch.Tensor,
            temperature: float,
            top_k: int,
            top_p: float,
            attention_mask: torch.Tensor,
            use_self_attn_cache: bool = True,
            init_step: bool = False,
    ) -> tuple[int, torch.Tensor, torch.Tensor]:
        device = input_ids.device

        # Forward pass to get next token logits
        outputs = self.forward(
            input_ids[:, -1].unsqueeze(0) if not init_step else input_ids, attention_mask=attention_mask[:, -1].unsqueeze(0) if not init_step else attention_mask,
            use_self_attn_cache=use_self_attn_cache, current_positions=torch.tensor([[input_ids.size(-1)]]).to(input_ids.device) if not init_step else None
        )
        next_token_logits = outputs[:, -1, :]  # Get logits for next token
        # Apply sampling
        next_token = sample(
            next_token_logits,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
        )
        next_token = next_token.item()  # Extract scalar token
        next_token_ten = torch.tensor([[next_token]], device=device)
        next_input_ids = torch.cat([input_ids, next_token_ten], dim=1)
        new_one = torch.ones(1, 1, dtype=torch.bool, device=device)
        next_mask = torch.cat([attention_mask, new_one], dim=1) if attention_mask is not None else None

        # Yield the generated token
        return (
            next_token,
            next_input_ids,
            next_mask
        )

    def generate(
            self,
            input_ids: torch.Tensor,
            attention_mask: torch.Tensor = None,
            temperature: float = 1.0,
            top_k: Optional[int] = None,
            top_p: Optional[float] = None,
            max_seq_len: int = 256,
            use_self_attn_cache: bool = True,
            answer_token_id: Optional[int] = 6,
            eos_token_id: Optional[int] = 3,
    ) -> Iterator[int]:
        assert input_ids.size(0) == 1, 'Batch size must be 1 in single interaction mode'

        with torch.no_grad():
            self.reset_self_attn_cache()

            input_ids = torch.cat([input_ids, torch.tensor([[answer_token_id]]).to(input_ids.device)], dim=-1)
            attention_mask = torch.cat([attention_mask, torch.ones(1, 1, device=attention_mask.device)], dim=-1)

            for i in range(max_seq_len - input_ids.size(-1)):
                next_token, input_ids, attention_mask = self._generate_single_token(
                    input_ids, temperature, top_k, top_p, attention_mask,
                    use_self_attn_cache=use_self_attn_cache, init_step=i == 0 or not use_self_attn_cache
                )
                yield next_token
                if next_token == eos_token_id:
                    break

    def stringify_token(self, tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast], token_id: int, skip_special_tokens: bool = True) -> str:
        return decode_post_process(tokenizer.decode([token_id], skip_special_tokens=skip_special_tokens))

    def _build_full_conversation_text(self, chat_history: list[list[str]], new_query: str, tokenizer_config: dict) -> str:
        full_text = f"{tokenizer_config['bos_token']}"

        for query, answer in chat_history:
            full_text += f"{tokenizer_config['query_token']}{query}{tokenizer_config['answer_token']}{answer}"

        full_text += f"{tokenizer_config['query_token']}{new_query}"
        return full_text

    def tokenize_chat_template(
            self,
            tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
            chat_history: list[Union[dict[Literal['role', 'content'], str], list[str]]],
            new_query: str,
            max_seq_len: int = 256,
            use_simplified_format: bool = False,
            tokenized_config: dict = None
    ) -> dict[str, torch.Tensor]:
            if not use_simplified_format:
                def get_simplified_item(idx: int) -> list[str]:
                    query = chat_history[idx]
                    assert query['role'] == 'user'
                    answer = chat_history[idx + 1]
                    assert answer['role'] == 'user'
                    return [query['content'], answer['content']]
                chat_history = [get_simplified_item(i) for i in range(0, len(chat_history), 2)]

            if tokenized_config is None:
                tokenized_config = {
                    'bos_token': '[BOS]',
                    'eos_token': '[EOS]',
                    'query_token': '[Q]',
                    'answer_token': '[A]',
                }

            full_text = self._build_full_conversation_text(chat_history, new_query, tokenized_config)

            enc = tokenizer(
                full_text,
                max_length=max_seq_len,
                truncation=True,
                padding=False,
                return_tensors='pt',
                return_attention_mask=True,
                add_special_tokens=False
            )

            input_ids = enc['input_ids']
            attention_mask = enc['attention_mask']

            return {
                'input_ids': input_ids,
                'attention_mask': attention_mask,
            }